import logging
from datetime import datetime

from ws4ever.views import BaseWebSocketApplication

log = logging.getLogger('WebSocketClientApplication')


class WebSocketClientApplication(BaseWebSocketApplication):
    def __init__(self, ws):
        meta = ws.environ
        log.info('%s - - [%s] "%s %s %s"',
                 meta.get("REMOTE_ADDR") or meta.get("HTTP_X_FORWARDED_FOR") or meta.get("HTTP_X_REAL_IP"),
                 datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                 meta.get('REQUEST_METHOD'),
                 meta.get('PATH_INFO'),
                 meta.get('SERVER_PROTOCOL'))

        super(WebSocketClientApplication, self).__init__(ws)

    def on_ping(self):
        log.debug('on ping')

    def handle_message(self, kind, payload=None):
        log.debug('got message kind:%s, payload:%s', kind, payload)
        self.send_message(kind, payload)

    def on_connect(self):
        log.debug('%s ... on connect', self)
        # import time
        # while True:
        #     time.sleep(1)

    def on_disconnect(self):
        log.debug('%s ... on disconnect', self)

    @classmethod
    def on_redis_message(cls, data):
        log.debug('[sub] <-- %s', data)



import redis
from django.conf import settings

redis_client = redis.StrictRedis(settings.REDIS_HOST, password=settings.REDIS_PASSWORD, decode_responses=True, db=0)
CH_WEBSOCKET_NOTIFY = 'websocket:notify'  # 发送广播


def remove_empty_keys(d):
    for k, v in list(d.items()):
        if v is None:
            del d[k]

# when you need to publish message, you can
# rc_main.publish(CH_WEBSOCKET_NOTIFY, json.dumps({'kind':'','payload':''}))



import importlib

from django.conf import settings


def get_handler_class(class_name):
    tokens = class_name.split('.')
    pkg = importlib.import_module('.'.join(tokens[:-1]))
    return getattr(pkg, tokens[-1])


def get_websocket_sources():
    return [(path, get_handler_class(handler_class))
            for path, handler_class in settings.WEBSOCKET_PATHS.items()]



import json
import logging
import time
import uuid
from typing import Dict, Optional
from urllib.parse import parse_qsl

from django.conf import settings
from geventwebsocket import WebSocketApplication
from ipware.ip import get_ip
from memoize import memoize
from voluptuous import Any

log = logging.getLogger('BaseWebSocketApplication')


class BaseWebSocketApplication(WebSocketApplication):
    clients = {}  # type:Dict[str: BaseWebSocketApplication]

    def __init__(self, ws):
        super(BaseWebSocketApplication, self).__init__(ws)

        self.META = self.ws.environ
        self.query = dict(parse_qsl(self.META.get('QUERY_STRING', '')))
        self.remote_ip = get_ip(self)
        self.guid = "{}|{}".format(self.remote_ip, uuid.uuid4())
        self.last_ping = None

    def on_open(self):
        self.last_ping = time.time()
        self.clients[self.guid] = self

        self.on_connect()
        self.check_zombie_channels()

    def handle_message(self, kind: str, payload: Optional[Any] = None):
        pass

    def on_ping(self):
        pass

    # called by on_open
    def on_connect(self):
        pass

    def on_disconnect(self):
        pass

    def send_message(self, kind: str, payload: Optional[Any] = None):
        obj = {'kind': kind}
        if payload is not None:
            obj['payload'] = payload
        self.write_message(json.dumps(obj))

    # internal =================
    def __repr__(self):
        return '[{}:{}]'.format(self.__class__.__name__, self.guid)

    def write_message(self, message):
        self.ws.send(message)

    def on_message(self, message, *args, **kwargs):
        if self.guid not in self.clients:
            self.close(500, 'connection died')

        self.last_ping = time.time()

        if message is None:
            # 可能来自客户端关闭, client会自动关闭,这里不处理
            return

        elif len(message) == 1:
            self.ws.send('.')  # pong back
            self.on_ping()

        else:
            try:
                obj = json.loads(message)
                kind, payload = map(obj.get, ['kind', 'payload'])
                self.handle_message(kind, payload)
            except Exception as e:
                log.warning('%s <-!(%s) %s', self, e, message)

    def on_close(self, *args, **kwargs):
        if self.guid:
            del self.clients[self.guid]
        self.on_disconnect()

    def close(self, *args, **kwargs):
        self.ws.close(*args, **kwargs)

    @classmethod
    @memoize(settings.CHECK_ZOMBIE_INTERVAL)
    def check_zombie_channels(cls):
        now = time.time()

        kills = 0
        clients = cls.clients.copy()
        for client in clients.values():
            dt = now - client.last_ping
            if dt > settings.CHECK_ZOMBIE_INTERVAL * 2:
                kills += 1
                client.close()
        log.debug('[%s] check ping time: [%s/-%s] @%.2f ms',
                  cls.__name__,
                  len(cls.clients),
                  kills,
                  (time.time() - now) * 1000)



"""
WSGI config for django_ws4ever project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/1.10/howto/deployment/wsgi/
"""

import os

from django.core.wsgi import get_wsgi_application

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "django_ws4ever.settings")
# from django_ws4ever.views import WebSocketClientApplication

application = get_wsgi_application()

# gunicorn 用的 websocket wsgi
from geventwebsocket import Resource
from ws4ever.helpers import get_websocket_sources
ws_application = Resource(get_websocket_sources())




# @formatter:off
from gevent import monkey;monkey.patch_all(thread=False)
# @formatter:on

import errno
import os
import socket
import sys
from datetime import datetime

import gevent
from django.conf import settings
from django.contrib.staticfiles.handlers import StaticFilesHandler
from django.core.management.commands.runserver import Command as RunserverCommand
from django.utils import autoreload
from django.utils.encoding import force_text
from geventwebsocket import Resource, WebSocketServer

from ws4ever.helpers import get_websocket_sources

def get_server(addr, port, wsgi_handler):
    resources = get_websocket_sources()
    resources.append(('^/.*', wsgi_handler))
    return WebSocketServer((addr, port), Resource(resources))


def run(addr, port, wsgi_handler):
    server = get_server(addr, port, wsgi_handler)
    g = gevent.spawn(server.serve_forever)
    gevent.joinall([g])


class Command(RunserverCommand):
    def add_arguments(self, parser):
        # remove threading from origin code
        parser.add_argument(
            'addrport', nargs='?',
            help='Optional port number, or ipaddr:port'
        )
        parser.add_argument(
            '--ipv6', '-6', action='store_true', dest='use_ipv6', default=False,
            help='Tells Django to use an IPv6 address.',
        )
        parser.add_argument(
            '--noreload', action='store_false', dest='use_reloader', default=True,
            help='Tells Django to NOT use the auto-reloader.',
        )

    def get_handler(self, *args, **options):
        # add static file handler
        handler = super(Command, self).get_handler(*args, **options)
        return StaticFilesHandler(handler)

    def inner_run(self, *args, **options):
        # origin code ----------------
        autoreload.raise_last_exception()

        # 'shutdown_message' is a stealth option.
        shutdown_message = options.get('shutdown_message', '')
        quit_command = 'CTRL-BREAK' if sys.platform == 'win32' else 'CONTROL-C'

        self.stdout.write("Performing system checks...\n\n")
        self.check(display_num_errors=True)

        # Need to check migrations here, so can't use the
        # requires_migrations_check attribute.
        self.check_migrations()
        now = datetime.now().strftime('%B %d, %Y - %X')
        self.stdout.write(now)
        self.stdout.write(
            (
                "Django version %(version)s, using settings %(settings)r\n"
                "Starting development server at http://%(addr)s:%(port)s/\n"
                "Quit the server with %(quit_command)s.\n"
            ) % {
                "version": self.get_version(),
                "settings": settings.SETTINGS_MODULE,
                "addr": '[%s]' % self.addr if self._raw_ipv6 else self.addr,
                "port": self.port,
                "quit_command": quit_command,
            }
        )
        # origin code ends ----------------
        try:
            run(self.addr, int(self.port), self.get_handler())

        # origin code to end -------------------
        except socket.error as e:
            # Use helpful error messages instead of ugly tracebacks.
            errors = {
                errno.EACCES: "You don't have permission to access that port.",
                errno.EADDRINUSE: "That port is already in use.",
                errno.EADDRNOTAVAIL: "That IP address can't be assigned to.",
            }
            try:
                error_text = errors[e.errno]
            except KeyError:
                error_text = force_text(e)
            self.stderr.write("Error: %s" % error_text)
            # Need to use an OS exit because sys.exit doesn't work in a thread
            os._exit(1)
        except KeyboardInterrupt:
            if shutdown_message:
                self.stdout.write(shutdown_message)
            sys.exit(0)



import importlib
import logging
import os
import gevent
from django.apps import AppConfig
from django.conf import settings
import redis
from django.utils.encoding import force_str

from ws4ever.redis import CH_WEBSOCKET_NOTIFY

log = logging.getLogger('ws4ever')


class Ws4EverConfig(AppConfig):
    name = 'ws4ever'

    def ready(self):
        self.add_period_task(self.subscribe_redis, 1)

    @classmethod
    def add_period_task(cls, func, interval: float):
        def exec_func():
            while True:
                gevent.sleep(interval)
                try:
                    func()
                except:
                    log.exception("exec fun exception %s", str(func))

        return gevent.spawn(exec_func)

    @classmethod
    def subscribe_redis(cls):
        log.info('start subscribe, pid:%s', os.getpid())
        r = redis.StrictRedis(settings.REDIS_HOST, password=settings.REDIS_PASSWORD)
        p = r.pubsub(ignore_subscribe_messages=True)
        p.psubscribe("*")
        for message in p.listen():
            cls.handle_redis_message(message)

    @classmethod
    def handle_redis_message(cls, m):
        # from main.websocket import WebSocketClientApplication
        try:
            channel, data = (force_str(m.get(x, b'')) for x in ['channel', 'data'])
            if channel == CH_WEBSOCKET_NOTIFY:
                for handlers in settings.WEBSOCKET_PATHS.values():
                    importlib.import_module(handlers).on_redis_message(data)
            else:
                log.warning("unexpected redis message channel: %s data: %s", channel, data)
        except:
            log.exception('handle_redis_message')
